@@ -1,6 +1,6 @@
 /*
     ccid.c: CCID common code
-    Copyright (C) 2003-2005   Ludovic Rousseau
+    Copyright (C) 2003-2010   Ludovic Rousseau
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -18,7 +18,7 @@
 */
 
 /*
- * $Id: ccid.c 4346 2009-07-28 13:39:37Z rousseau $
+ * $Id: ccid.c 5182 2010-08-24 09:46:45Z rousseau $
  */
 
 #include <stdio.h>
@@ -33,7 +33,10 @@
 #include "defs.h"
 #include "ccid_ifdhandler.h"
 #include "commands.h"
-#include "ccid_usb.h"
+
+#ifdef __APPLE__
+#include <CoreFoundation/CoreFoundation.h>
+#endif
 
 /*****************************************************************************
  *
@@ -43,7 +46,6 @@
 int ccid_open_hack_pre(unsigned int reader_index)
 {
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
-	int doInterruptRead = 1;
 
 	switch (ccid_descriptor->readerID)
 	{
@@ -60,26 +62,22 @@ int ccid_open_hack_pre(unsigned int reader_index)
 		case CL1356D:
 			/* the firmware needs some time to initialize */
 			(void)sleep(1);
-			ccid_descriptor->readTimeout = 60; /* 60 seconds */
-			break;
-
-		case KOBIL_TRIBANK:
-			/* the InterruptRead does not timeout (on Mac OS X) */
-			doInterruptRead = 0;
+			ccid_descriptor->readTimeout = 60*1000; /* 60 seconds */
 			break;
 	}
 
 	/* CCID */
-	if (doInterruptRead && (0 == ccid_descriptor->bInterfaceProtocol))
+	if ((PROTOCOL_CCID == ccid_descriptor->bInterfaceProtocol)
+		&& (3 == ccid_descriptor -> bNumEndpoints))
 	{
 #ifndef TWIN_SERIAL
-		/* just wait for 10ms in case a notification is in the pipe */
-		(void)InterruptRead(reader_index, 10);
+		/* just wait for 100ms in case a notification is in the pipe */
+		(void)InterruptRead(reader_index, 100);
 #endif
 	}
 
 	/* ICCD type A */
-	if (ICCD_A == ccid_descriptor->bInterfaceProtocol)
+	if (PROTOCOL_ICCD_A == ccid_descriptor->bInterfaceProtocol)
 	{
 		unsigned char tmp[MAX_ATR_SIZE];
 		unsigned int n = sizeof(tmp);
@@ -91,7 +89,7 @@ int ccid_open_hack_pre(unsigned int reader_index)
 	}
 
 	/* ICCD type B */
-	if (ICCD_B == ccid_descriptor->bInterfaceProtocol)
+	if (PROTOCOL_ICCD_B == ccid_descriptor->bInterfaceProtocol)
 	{
 		unsigned char tmp[MAX_ATR_SIZE];
 		unsigned int n = sizeof(tmp);
@@ -126,10 +124,10 @@ int ccid_open_hack_post(unsigned int reader_index)
 	{
 		case GEMPCKEY:
 		case GEMPCTWIN:
-			/* Reader announces TPDU but can do APDU */
+			/* Reader announces TPDU but can do APDU (EMV in fact) */
 			if (DriverOptions & DRIVER_OPTION_GEMPC_TWIN_KEY_APDU)
 			{
-				unsigned char cmd[] = { 0xA0, 0x02 };
+				unsigned char cmd[] = { 0x1F, 0x02 };
 				unsigned char res[10];
 				unsigned int length_res = sizeof(res);
 
@@ -141,6 +139,7 @@ int ccid_open_hack_post(unsigned int reader_index)
 			}
 			break;
 
+		case VEGAALPHA:
 		case GEMPCPINPAD:
 			/* load the l10n strings in the pinpad memory */
 			{
@@ -249,10 +248,28 @@ int ccid_open_hack_post(unsigned int reader_index)
 					"Card Error",
 					"PIN blocked" };
 
-				char *lang;
+				const char *lang;
 				const char **l10n;
 
+#ifdef __APPLE__
+				CFArrayRef cfa;
+				CFStringRef slang;
+
+				/* Get the complete ordered list */
+				cfa = CFLocaleCopyPreferredLanguages();
+
+				/* Use the first/preferred language
+				 * As the driver is run as root we get the language
+				 * selected during install */
+				slang = CFArrayGetValueAtIndex(cfa, 0);
+
+				/* CFString -> C string */
+				lang = CFStringGetCStringPtr(slang, kCFStringEncodingMacRoman);
+#else
+				/* The other Unixes just use the LANG env variable */
 				lang = getenv("LANG");
+#endif
+				DEBUG_COMM2("Using lang: %s", lang);
 				if (NULL == lang)
 					l10n = en;
 				else
@@ -275,6 +292,10 @@ int ccid_open_hack_post(unsigned int reader_index)
 						l10n = en;
 				}
 
+#ifdef __APPLE__
+				/* Release the allocated array */
+				CFRelease(cfa);
+#endif
 				offset = 0;
 				cmd[offset++] = 0xB2;	/* load strings */
 				cmd[offset++] = 0xA0;	/* address of the memory */
