@@ -1,6 +1,6 @@
 /*
     ccid.h: CCID structures
-    Copyright (C) 2003-2009   Ludovic Rousseau
+    Copyright (C) 2003-2010   Ludovic Rousseau
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -18,7 +18,7 @@
 */
 
 /*
- * $Id: ccid.h 4280 2009-06-26 14:58:23Z rousseau $
+ * $Id: ccid.h 5475 2011-01-04 09:54:34Z rousseau $
  */
 
 typedef struct
@@ -86,7 +86,7 @@ typedef struct
 
 	/*
 	 * Read communication port timeout
-	 * value is seconds
+	 * value is milliseconds
 	 * this value can evolve dynamically if card request it (time processing).
 	 */
 	unsigned int readTimeout;
@@ -121,6 +121,11 @@ typedef struct
 	 * 4 = 1.8V
 	 */
 	int bVoltageSupport;
+
+	/*
+	 * USB serial number of the device (if any)
+	 */
+	char *sIFD_serial_number;
 } _ccid_descriptor;
 
 /* Features from dwFeatures */
@@ -150,8 +155,9 @@ typedef struct
 #define CCID_TIME_EXTENSION			0x80	/* 10 0000 00 */
 
 /* bInterfaceProtocol for ICCD */
-#define ICCD_A	1	/* ICCD Version A */
-#define ICCD_B	2	/* ICCD Version B */
+#define PROTOCOL_CCID	0	/* plain CCID */
+#define PROTOCOL_ICCD_A	1	/* ICCD Version A */
+#define PROTOCOL_ICCD_B	2	/* ICCD Version B */
 
 /* Product identification for special treatments */
 #define GEMPC433	0x08E64433
@@ -179,7 +185,13 @@ typedef struct
 #define SEG			0x08E68000
 #define BLUDRIVEII_CCID	0x1B0E1078
 #define DELLSCRK    0x413C2101
+#define DELLSK      0x413C2100
 #define KOBIL_TRIBANK	0x0D463010
+#define KOBIL_MIDENTITY_VISUAL	0x0D460D46
+#define VEGAALPHA   0x09820008
+
+#define VENDOR_GEMALTO 0x08E6
+#define GET_VENDOR(readerID) ((readerID >> 16) & 0xFFFF)
 
 /*
  * The O2Micro OZ776S reader has a wrong USB descriptor
