@@ -1,6 +1,6 @@
 /*
  * ccid_serial.c: communicate with a GemPC Twin smart card reader
- * Copyright (C) 2001-2004 Ludovic Rousseau <ludovic.rousseau@free.fr>
+ * Copyright (C) 2001-2010 Ludovic Rousseau <ludovic.rousseau@free.fr>
  *
  * Thanks to Niki W. Waibel <niki.waibel@gmx.net> for a prototype version
  *
@@ -20,7 +20,7 @@
  */
 
 /*
- * $Id: ccid_serial.c 3488 2009-05-07 08:04:20Z rousseau $
+ * $Id: ccid_serial.c 5382 2010-11-05 22:49:01Z rousseau $
  */
 
 #include <stdio.h>
@@ -35,9 +35,9 @@
 #include <sys/ioctl.h>
 #include <ifdhandler.h>
 
+#include "config.h"
 #include "defs.h"
 #include "ccid_ifdhandler.h"
-#include "config.h"
 #include "debug.h"
 #include "ccid.h"
 #include "utils.h"
@@ -310,6 +310,12 @@ ack:
 	/* total frame size */
 	to_read = 10+dw2i(buffer, 1);
 
+	if ((to_read < 10) || (to_read > (int)*length))
+	{
+		DEBUG_CRITICAL2("Wrong value for frame size: %d", to_read);
+		return STATUS_COMM_ERROR;
+	}
+
 	DEBUG_COMM2("frame size: %d", to_read);
 	if ((rv = get_bytes(reader_index, buffer+5, to_read-5)) != STATUS_SUCCESS)
 		return rv;
@@ -421,8 +427,8 @@ static int ReadChunk(unsigned int reader_index, unsigned char *buffer,
 		/* use select() to, eventually, timeout */
 		FD_ZERO(&fdset);
 		FD_SET(fd, &fdset);
-		t.tv_sec = serialDevice[reader_index].ccid.readTimeout;
-		t.tv_usec = 0;
+		t.tv_sec = serialDevice[reader_index].ccid.readTimeout / 1000;
+		t.tv_usec = (serialDevice[reader_index].ccid.readTimeout - t.tv_sec*1000)*1000;
 
 		i = select(fd+1, &fdset, NULL, NULL, &t);
 		if (i == -1)
@@ -433,7 +439,7 @@ static int ReadChunk(unsigned int reader_index, unsigned char *buffer,
 		else
 			if (i == 0)
 			{
-				DEBUG_COMM2("Timeout! (%d sec)", serialDevice[reader_index].ccid.readTimeout);
+				DEBUG_COMM2("Timeout! (%d ms)", serialDevice[reader_index].ccid.readTimeout);
 				return -1;
 			}
 
@@ -544,6 +550,7 @@ static status_t set_ccid_descriptor(unsigned int reader_index,
 
 			*serialDevice[reader_index].nb_opened_slots += 1;
 			serialDevice[reader_index].ccid.bCurrentSlotIndex++;
+			serialDevice[reader_index].ccid.dwSlotStatus = IFD_ICC_PRESENT;
 			DEBUG_INFO2("Opening slot: %d",
 					serialDevice[reader_index].ccid.bCurrentSlotIndex);
 			switch (readerID)
@@ -640,7 +647,8 @@ status_t OpenSerialByName(unsigned int reader_index, char *dev_name)
 {
 	struct termios current_termios;
 	unsigned int reader = reader_index;
-	char reader_name[TOKEN_MAX_VALUE_SIZE] = "GemPCTwin";
+	/* 255 is MAX_DEVICENAME in pcscd.h */
+	char reader_name[255] = "GemPCTwin";
 	char *p;
 	status_t ret;
 
@@ -741,7 +749,7 @@ status_t OpenSerialByName(unsigned int reader_index, char *dev_name)
 		return STATUS_UNSUCCESSFUL;
 	}
 
-	/* perform a command to be sure a Gemplus reader is connected
+	/* perform a command to be sure a Gemalto reader is connected
 	 * get the reader firmware */
 	{
 		unsigned char tx_buffer[] = { 0x02 };
@@ -749,7 +757,7 @@ status_t OpenSerialByName(unsigned int reader_index, char *dev_name)
 		unsigned int rx_length = sizeof(rx_buffer);
 
 		/* 2 seconds timeout to not wait too long if no reader is connected */
-		serialDevice[reader].ccid.readTimeout = 2;
+		serialDevice[reader].ccid.readTimeout = 2*1000;
 
 		if (IFD_SUCCESS != CmdEscape(reader_index, tx_buffer, sizeof(tx_buffer),
 			rx_buffer, &rx_length))
@@ -783,6 +791,8 @@ status_t OpenSerialByName(unsigned int reader_index, char *dev_name)
 		}
 	}
 
+	serialDevice[reader_index].ccid.sIFD_serial_number = NULL;
+
 	return STATUS_SUCCESS;
 } /* OpenSerialByName */
 
