@@ -1,6 +1,6 @@
 /*
     ifdhandler.c: IFDH API
-    Copyright (C) 2003-2009   Ludovic Rousseau
+    Copyright (C) 2003-2010   Ludovic Rousseau
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -17,14 +17,14 @@
 	Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
-/* $Id: ifdhandler.c 4346 2009-07-28 13:39:37Z rousseau $ */
+/* $Id: ifdhandler.c 5515 2011-01-18 15:18:29Z rousseau $ */
 
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
 #include <arpa/inet.h>
-#include "misc.h"
 #include "config.h"
+#include "misc.h"
 #include <pcsclite.h>
 #include <ifdhandler.h>
 #include <reader.h>
@@ -57,10 +57,6 @@ int PowerOnVoltage = VOLTAGE_5V;
 static int DebugInitialized = FALSE;
 
 /* local functions */
-#if HAVE_DECL_TAG_IFD_POLLING_THREAD && !defined(TWIN_SERIAL) && defined(USE_USB_INTERRUPT)
-static RESPONSECODE IFDHPolling(DWORD Lun);
-static RESPONSECODE IFDHSleep(DWORD Lun);
-#endif
 static void init_driver(void);
 static void extra_egt(ATR_t *atr, _ccid_descriptor *ccid_desc, DWORD Protocol);
 static char find_baud_rate(unsigned int baudrate, unsigned int *list);
@@ -68,6 +64,7 @@ static unsigned int T0_card_timeout(double f, double d, int TC1, int TC2,
 	int clock_frequency);
 static unsigned int T1_card_timeout(double f, double d, int TC1, int BWI,
 	int CWI, int clock_frequency);
+static int get_IFSC(ATR_t *atr, int *i);
 
 
 EXTERNAL RESPONSECODE IFDHCreateChannelByName(DWORD Lun, LPSTR lpcDevice)
@@ -112,17 +109,27 @@ EXTERNAL RESPONSECODE IFDHCreateChannelByName(DWORD Lun, LPSTR lpcDevice)
 	}
 	else
 	{
+		unsigned char pcbuffer[SIZE_GET_SLOT_STATUS];
+		unsigned int oldReadTimeout;
+		_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
+
 		/* Maybe we have a special treatment for this reader */
 		(void)ccid_open_hack_pre(reader_index);
 
+		/* save the current read timeout computed from card capabilities */
+		oldReadTimeout = ccid_descriptor->readTimeout;
+
+		/* 1000ms just to resync the USB toggle bits */
+		ccid_descriptor->readTimeout = 1000;
+
 		/* Try to access the reader */
 		/* This "warm up" sequence is sometimes needed when pcscd is
 		 * restarted with the reader already connected. We get some
 		 * "usb_bulk_read: Resource temporarily unavailable" on the first
 		 * few tries. It is an empirical hack */
-		if ((IFD_COMMUNICATION_ERROR == IFDHICCPresence(Lun))
-			&& (IFD_COMMUNICATION_ERROR == IFDHICCPresence(Lun))
-			&& (IFD_COMMUNICATION_ERROR == IFDHICCPresence(Lun)))
+		if ((IFD_COMMUNICATION_ERROR == CmdGetSlotStatus(reader_index, pcbuffer))
+			&& (IFD_COMMUNICATION_ERROR == CmdGetSlotStatus(reader_index, pcbuffer))
+			&& (IFD_COMMUNICATION_ERROR == CmdGetSlotStatus(reader_index, pcbuffer)))
 		{
 			DEBUG_CRITICAL("failed");
 			return_value = IFD_COMMUNICATION_ERROR;
@@ -131,9 +138,14 @@ EXTERNAL RESPONSECODE IFDHCreateChannelByName(DWORD Lun, LPSTR lpcDevice)
 			(void)ClosePort(reader_index);
 			ReleaseReaderIndex(reader_index);
 		}
-		else 
+		else
+		{
+			/* set back the old timeout */
+			ccid_descriptor->readTimeout = oldReadTimeout;
+
 			/* Maybe we have a special treatment for this reader */
 			(void)ccid_open_hack_post(reader_index);
+		}
 	}
 
 #ifdef HAVE_PTHREAD
@@ -214,17 +226,27 @@ EXTERNAL RESPONSECODE IFDHCreateChannel(DWORD Lun, DWORD Channel)
 	}
 	else
 	{
+		unsigned char pcbuffer[SIZE_GET_SLOT_STATUS];
+		unsigned int oldReadTimeout;
+		_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
+
 		/* Maybe we have a special treatment for this reader */
 		(void)ccid_open_hack_pre(reader_index);
 
+		/* save the current read timeout computed from card capabilities */
+		oldReadTimeout = ccid_descriptor->readTimeout;
+
+		/* 100ms just to resync the USB toggle bits */
+		ccid_descriptor->readTimeout = 100;
+
 		/* Try to access the reader */
 		/* This "warm up" sequence is sometimes needed when pcscd is
 		 * restarted with the reader already connected. We get some
 		 * "usb_bulk_read: Resource temporarily unavailable" on the first
 		 * few tries. It is an empirical hack */
-		if ((IFD_COMMUNICATION_ERROR == IFDHICCPresence(Lun))
-			&& (IFD_COMMUNICATION_ERROR == IFDHICCPresence(Lun))
-			&& (IFD_COMMUNICATION_ERROR == IFDHICCPresence(Lun)))
+		if ((IFD_COMMUNICATION_ERROR == CmdGetSlotStatus(reader_index, pcbuffer))
+			&& (IFD_COMMUNICATION_ERROR == CmdGetSlotStatus(reader_index, pcbuffer))
+			&& (IFD_COMMUNICATION_ERROR == CmdGetSlotStatus(reader_index, pcbuffer)))
 		{
 			DEBUG_CRITICAL("failed");
 			return_value = IFD_COMMUNICATION_ERROR;
@@ -234,8 +256,13 @@ EXTERNAL RESPONSECODE IFDHCreateChannel(DWORD Lun, DWORD Channel)
 			ReleaseReaderIndex(reader_index);
 		}
 		else
+		{
+			/* set back the old timeout */
+			ccid_descriptor->readTimeout = oldReadTimeout;
+
 			/* Maybe we have a special treatment for this reader */
 			(void)ccid_open_hack_post(reader_index);
+		}
 	}
 
 #ifdef HAVE_PTHREAD
@@ -290,37 +317,33 @@ EXTERNAL RESPONSECODE IFDHCloseChannel(DWORD Lun)
 } /* IFDHCloseChannel */
 
 
-#if HAVE_DECL_TAG_IFD_POLLING_THREAD && !defined(TWIN_SERIAL) && defined(USE_USB_INTERRUPT)
-static RESPONSECODE IFDHPolling(DWORD Lun)
+#if !defined(TWIN_SERIAL)
+static RESPONSECODE IFDHPolling(DWORD Lun, int timeout)
 {
 	int reader_index;
-	int ret;
 
 	if (-1 == (reader_index = LunToReaderIndex(Lun)))
 		return IFD_COMMUNICATION_ERROR;
 
 	/* log only if DEBUG_LEVEL_PERIODIC is set */
 	if (LogLevel & DEBUG_LEVEL_PERIODIC)
-		DEBUG_INFO3("%s (lun: %X)", CcidSlots[reader_index].readerName, Lun);
+		DEBUG_INFO4("%s (lun: %X) %d ms", CcidSlots[reader_index].readerName,
+			Lun, timeout);
 
-	ret = InterruptRead(reader_index, 2*1000);	/* 2 seconds */
-	if (ret > 0)
-		return IFD_SUCCESS;
-	if (0 == ret)
-		return IFD_NO_SUCH_DEVICE;
-	return IFD_COMMUNICATION_ERROR;
+	return InterruptRead(reader_index, timeout);
 }
 
 /* on an ICCD device the card is always inserted
  * so no card movement will ever happen: just do nothing */
-static RESPONSECODE IFDHSleep(DWORD Lun)
+static RESPONSECODE IFDHSleep(DWORD Lun, int timeout)
 {
 	int reader_index;
 
 	if (-1 == (reader_index = LunToReaderIndex(Lun)))
 		return IFD_COMMUNICATION_ERROR;
 
-	DEBUG_INFO3("%s (lun: %X)", CcidSlots[reader_index].readerName, Lun);
+	DEBUG_INFO4("%s (lun: %X) %d ms", CcidSlots[reader_index].readerName, Lun,
+		timeout);
 
 	/* just sleep for 5 seconds since the polling thread is NOT killable
 	 * so pcscd event thread must loop to exit cleanly
@@ -329,7 +352,20 @@ static RESPONSECODE IFDHSleep(DWORD Lun)
 	 * TAG_IFD_POLLING_THREAD_KILLABLE then we could use a much longer delay
 	 * and be killed before pcscd exits
 	 */
-	(void)sleep(600);	/* 10 minutes */
+	(void)usleep(timeout);
+	return IFD_SUCCESS;
+}
+
+static RESPONSECODE IFDHStopPolling(DWORD Lun)
+{
+	int reader_index;
+
+	if (-1 == (reader_index = LunToReaderIndex(Lun)))
+		return IFD_COMMUNICATION_ERROR;
+
+	DEBUG_INFO3("%s (lun: %X)", CcidSlots[reader_index].readerName, Lun);
+
+	(void)InterruptStop(reader_index);
 	return IFD_SUCCESS;
 }
 #endif
@@ -356,6 +392,7 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 	 * IFD_SUCCESS IFD_ERROR_TAG
 	 */
 	int reader_index;
+	RESPONSECODE return_value = IFD_SUCCESS;
 
 	if (-1 == (reader_index = LunToReaderIndex(Lun)))
 		return IFD_COMMUNICATION_ERROR;
@@ -370,11 +407,41 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 			/* If Length is not zero, powerICC has been performed.
 			 * Otherwise, return NULL pointer
 			 * Buffer size is stored in *Length */
-			*Length = (*Length < CcidSlots[reader_index].nATRLength) ?
-				*Length : CcidSlots[reader_index].nATRLength;
+			if ((int)*Length >= CcidSlots[reader_index].nATRLength)
+			{
+				*Length = CcidSlots[reader_index].nATRLength;
 
-			if (*Length)
 				memcpy(Value, CcidSlots[reader_index].pcATRBuffer, *Length);
+			}
+			else
+				return_value = IFD_ERROR_INSUFFICIENT_BUFFER;
+			break;
+
+		case SCARD_ATTR_ICC_INTERFACE_STATUS:
+			*Length = 1;
+			if (IFD_ICC_PRESENT == IFDHICCPresence(Lun))
+				/* nonzero if contact is active */
+				*Value = 1;
+			else
+				/* smart card electrical contact is not active */
+				*Value = 0;
+			break;
+
+		case SCARD_ATTR_ICC_PRESENCE:
+			*Length = 1;
+			/* Single byte indicating smart card presence:
+			 * 0 = not present
+			 * 1 = card present but not swallowed (applies only if
+			 *     reader supports smart card swallowing)
+			 * 2 = card present (and swallowed if reader supports smart
+			 *     card swallowing)
+			 * 4 = card confiscated. */
+			if (IFD_ICC_PRESENT == IFDHICCPresence(Lun))
+				/* Card present */
+				*Value = 2;
+			else
+				/* Not present */
+				*Value = 0;
 			break;
 
 #ifdef HAVE_PTHREAD
@@ -384,6 +451,8 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 				*Length = 1;
 				*Value = CCID_DRIVER_MAX_READERS;
 			}
+			else
+				return_value = IFD_ERROR_INSUFFICIENT_BUFFER;
 			break;
 
 		case TAG_IFD_THREAD_SAFE:
@@ -396,6 +465,8 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 				*Value = 1; /* Can talk to multiple readers at the same time */
 #endif
 			}
+			else
+				return_value = IFD_ERROR_INSUFFICIENT_BUFFER;
 			break;
 #endif
 
@@ -417,6 +488,8 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 #endif
 				DEBUG_INFO2("Reader supports %d slot(s)", *Value);
 			}
+			else
+				return_value = IFD_ERROR_INSUFFICIENT_BUFFER;
 			break;
 
 		case TAG_IFD_SLOT_THREAD_SAFE:
@@ -425,6 +498,8 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 				*Length = 1;
 				*Value = 0; /* Can NOT talk to multiple slots at the same time */
 			}
+			else
+				return_value = IFD_ERROR_INSUFFICIENT_BUFFER;
 			break;
 
 		case SCARD_ATTR_VENDOR_IFD_VERSION:
@@ -438,9 +513,14 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 
 		case SCARD_ATTR_VENDOR_NAME:
 #define VENDOR_NAME "Ludovic Rousseau"
-			*Length = sizeof(VENDOR_NAME);
-			if (Value)
-				memcpy(Value, VENDOR_NAME, sizeof(VENDOR_NAME));
+			if (*Length >= sizeof(VENDOR_NAME))
+			{
+				*Length = sizeof(VENDOR_NAME);
+				if (Value)
+					memcpy(Value, VENDOR_NAME, sizeof(VENDOR_NAME));
+			}
+			else
+				return_value = IFD_ERROR_INSUFFICIENT_BUFFER;
 			break;
 
 		case SCARD_ATTR_MAXINPUT:
@@ -449,8 +529,8 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 				*(uint32_t *)Value = get_ccid_descriptor(reader_index) -> dwMaxCCIDMessageLength -10;
 			break;
 
-#if HAVE_DECL_TAG_IFD_POLLING_THREAD && !defined(TWIN_SERIAL) && defined(USE_USB_INTERRUPT)
-		case TAG_IFD_POLLING_THREAD:
+#if !defined(TWIN_SERIAL)
+		case TAG_IFD_POLLING_THREAD_WITH_TIMEOUT:
 			{
 				_ccid_descriptor *ccid_desc;
 
@@ -458,8 +538,13 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 				*Length = 0;
 
 				ccid_desc = get_ccid_descriptor(reader_index);
+
+				/* more than one slot is not supported */
+				if (ccid_desc -> bMaxSlotIndex > 0)
+					break;
+
 				/* CCID and not ICCD */
-				if ((0 == ccid_desc -> bInterfaceProtocol)
+				if ((PROTOCOL_CCID == ccid_desc -> bInterfaceProtocol)
 					/* 3 end points */
 					&& (3 == ccid_desc -> bNumEndpoints))
 				{
@@ -468,8 +553,8 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 						*(void **)Value = IFDHPolling;
 				}
 
-				if ((ICCD_A == ccid_desc->bInterfaceProtocol)
-					|| (ICCD_B == ccid_desc->bInterfaceProtocol))
+				if ((PROTOCOL_ICCD_A == ccid_desc->bInterfaceProtocol)
+					|| (PROTOCOL_ICCD_B == ccid_desc->bInterfaceProtocol))
 				{
 					*Length = sizeof(void *);
 					if (Value)
@@ -486,8 +571,8 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 				*Length = 0;
 
 				ccid_desc = get_ccid_descriptor(reader_index);
-				if ((ICCD_A == ccid_desc->bInterfaceProtocol)
-					|| (ICCD_B == ccid_desc->bInterfaceProtocol))
+				if ((PROTOCOL_ICCD_A == ccid_desc->bInterfaceProtocol)
+					|| (PROTOCOL_ICCD_B == ccid_desc->bInterfaceProtocol))
 				{
 					*Length = 1;	/* 1 char */
 					if (Value)
@@ -495,13 +580,49 @@ EXTERNAL RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
 				}
 			}
 			break;
+
+		case TAG_IFD_STOP_POLLING_THREAD:
+			{
+				_ccid_descriptor *ccid_desc;
+
+				/* default value: not supported */
+				*Length = 0;
+
+				ccid_desc = get_ccid_descriptor(reader_index);
+				/* CCID and not ICCD */
+				if ((PROTOCOL_CCID == ccid_desc -> bInterfaceProtocol)
+					/* 3 end points */
+					&& (3 == ccid_desc -> bNumEndpoints))
+				{
+					*Length = sizeof(void *);
+					if (Value)
+						*(void **)Value = IFDHStopPolling;
+				}
+			}
+			break;
 #endif
 
+		case SCARD_ATTR_VENDOR_IFD_SERIAL_NO:
+			{
+				_ccid_descriptor *ccid_desc;
+
+				/* default value: not supported */
+				*Length = 0;
+
+				ccid_desc = get_ccid_descriptor(reader_index);
+				if (ccid_desc->sIFD_serial_number)
+				{
+					strcpy((char *)Value, ccid_desc->sIFD_serial_number);
+					*Length = strlen((char *)Value);
+				}
+			}
+			break;
+
 		default:
-			return IFD_ERROR_TAG;
+			return_value = IFD_ERROR_TAG;
 	}
 
-	return IFD_SUCCESS;
+	return return_value;
 } /* IFDHGetCapabilities */
 
 
@@ -525,7 +646,8 @@ EXTERNAL RESPONSECODE IFDHSetCapabilities(DWORD Lun, DWORD Tag,
 	 * IFD_ERROR_VALUE_READ_ONLY
 	 */
 
-	/* By default, say it worked */
+	(void)Length;
+	(void)Value;
 
 	int reader_index;
 
@@ -535,9 +657,6 @@ EXTERNAL RESPONSECODE IFDHSetCapabilities(DWORD Lun, DWORD Tag,
 	DEBUG_INFO4("tag: 0x%X, %s (lun: %X)", Tag,
 		CcidSlots[reader_index].readerName, Lun);
 
-	/* if (CheckLun(Lun))
-		return IFD_COMMUNICATION_ERROR; */
-
 	return IFD_NOT_SUPPORTED;
 } /* IFDHSetCapabilities */
 
@@ -833,6 +952,7 @@ EXTERNAL RESPONSECODE IFDHSetProtocolParameters(DWORD Lun, DWORD Protocol,
 		t1_state_t *t1 = &(ccid_slot -> t1);
 		RESPONSECODE ret;
 		double f, d;
+		int ifsc;
 
 		/* TA1 is not default */
 		if (PPS_HAS_PPS1(pps))
@@ -860,9 +980,9 @@ EXTERNAL RESPONSECODE IFDHSetProtocolParameters(DWORD Lun, DWORD Protocol,
 
 				{
 					/* Hack for OpenPGP card */
-					char openpgp_atr[] = { 0x3B, 0xFA, 0x13, 0x00, 0xFF, 0x81,
-						0x31, 0x80, 0x45, 0x00, 0x31, 0xC1, 0x73, 0xC0,
-						0x01, 0x00, 0x00, 0x90, 0x00, 0xB1 };
+					unsigned char openpgp_atr[] = { 0x3B, 0xFA, 0x13,
+						0x00, 0xFF, 0x81, 0x31, 0x80, 0x45, 0x00, 0x31,
+						0xC1, 0x73, 0xC0, 0x01, 0x00, 0x00, 0x90, 0x00, 0xB1 };
 
 					if (0 == memcmp(ccid_slot->pcATRBuffer, openpgp_atr,
 						ccid_slot->nATRLength))
@@ -886,19 +1006,14 @@ EXTERNAL RESPONSECODE IFDHSetProtocolParameters(DWORD Lun, DWORD Protocol,
 			(param[3] & 0xF0) >> 4 /* BWI */, param[3] & 0x0F /* CWI */,
 			ccid_desc->dwDefaultClock);
 
-		/* TAi (i>2) present? IFSC */
-		for (i=2; i<ATR_MAX_PROTOCOLS; i++)
-			if (atr.ib[i][ATR_INTERFACE_BYTE_TA].present)
-			{
-				DEBUG_COMM3("IFSC (TA%d) present: %d", i+1,
-					atr.ib[i][ATR_INTERFACE_BYTE_TA].value);
-				 param[5] = atr.ib[i][ATR_INTERFACE_BYTE_TA].value;
+		ifsc = get_IFSC(&atr, &i);
+		if (ifsc > 0)
+		{
+			DEBUG_COMM3("IFSC (TA%d) present: %d", i, ifsc);
+			param[5] = ifsc;
+		}
 
-				/* only the first TAi (i>2) must be used */
-				break;
-			}
-
-		DEBUG_COMM2("Timeout: %d seconds", ccid_desc->readTimeout);
+		DEBUG_COMM2("Timeout: %d ms", ccid_desc->readTimeout);
 
 		ret = SetParameters(reader_index, 1, sizeof(param), param);
 		if (IFD_SUCCESS != ret)
@@ -939,8 +1054,7 @@ EXTERNAL RESPONSECODE IFDHSetProtocolParameters(DWORD Lun, DWORD Protocol,
 		ccid_desc->readTimeout = T0_card_timeout(f, d, param[2] /* TC1 */,
 			param[3] /* TC2 */, ccid_desc->dwDefaultClock);
 
-		DEBUG_COMM2("Communication timeout: %d seconds",
-			ccid_desc->readTimeout);
+		DEBUG_COMM2("Communication timeout: %d ms", ccid_desc->readTimeout);
 
 		ret = SetParameters(reader_index, 0, sizeof(param), param);
 		if (IFD_SUCCESS != ret)
@@ -951,20 +1065,14 @@ EXTERNAL RESPONSECODE IFDHSetProtocolParameters(DWORD Lun, DWORD Protocol,
 	if (SCARD_PROTOCOL_T1 == Protocol)
 	{
 		t1_state_t *t1 = &(ccid_slot -> t1);
-		int i;
+		int i, ifsc;
 
-		/* TAi (i>2) present? */
-		for (i=2; i<ATR_MAX_PROTOCOLS; i++)
-			if (atr.ib[i][ATR_INTERFACE_BYTE_TA].present)
-			{
-				DEBUG_COMM3("IFSC (TA%d) present: %d", i+1,
-					atr.ib[i][ATR_INTERFACE_BYTE_TA].value);
-				(void)t1_set_param(t1, IFD_PROTOCOL_T1_IFSC,
-					atr.ib[i][ATR_INTERFACE_BYTE_TA].value);
-
-				/* only the first TAi (i>2) must be used */
-				break;
-			}
+		ifsc = get_IFSC(&atr, &i);
+		if (ifsc > 0)
+		{
+			DEBUG_COMM3("IFSC (TA%d) present: %d", i, ifsc);
+			(void)t1_set_param(t1, IFD_PROTOCOL_T1_IFSC, ifsc);
+		}
 
 		/* IFSD not negociated by the reader? */
 		if (! (ccid_desc->dwFeatures & CCID_CLASS_AUTO_IFSD))
@@ -1027,7 +1135,7 @@ EXTERNAL RESPONSECODE IFDHPowerICC(DWORD Lun, DWORD Action,
 
 	unsigned int nlength;
 	RESPONSECODE return_value = IFD_SUCCESS;
-	unsigned char pcbuffer[RESP_BUF_SIZE];
+	unsigned char pcbuffer[10+MAX_ATR_SIZE];
 	int reader_index;
 	const char *actions[] = { "PowerUp", "PowerDown", "Reset" };
 	unsigned int oldReadTimeout;
@@ -1078,7 +1186,7 @@ EXTERNAL RESPONSECODE IFDHPowerICC(DWORD Lun, DWORD Action,
 			 * 1 ETU = 372 cycles during ATR
 			 * with a 4 MHz clock => 29 seconds
 			 */
-			ccid_descriptor->readTimeout = 60;
+			ccid_descriptor->readTimeout = 60*1000;
 
 			nlength = sizeof(pcbuffer);
 			return_value = CmdPowerOn(reader_index, &nlength, pcbuffer,
@@ -1091,7 +1199,7 @@ EXTERNAL RESPONSECODE IFDHPowerICC(DWORD Lun, DWORD Action,
 			{
 				/* used by GemCore SIM PRO: no card is present */
 				get_ccid_descriptor(reader_index)->dwSlotStatus
-					= IFD_ICC_NOT_PRESENT;;
+					= IFD_ICC_NOT_PRESENT;
 
 				DEBUG_CRITICAL("PowerUp failed");
 				return_value = IFD_ERROR_POWER_ACTION;
@@ -1164,6 +1272,8 @@ EXTERNAL RESPONSECODE IFDHTransmitToICC(DWORD Lun, SCARD_IO_HEADER SendPci,
 	unsigned int rx_length;
 	int reader_index;
 
+	(void)RecvPci;
+
 	if (-1 == (reader_index = LunToReaderIndex(Lun)))
 		return IFD_COMMUNICATION_ERROR;
 
@@ -1199,12 +1309,15 @@ EXTERNAL RESPONSECODE IFDHControl(DWORD Lun, DWORD dwControlCode,
 	 *
 	 * Notes: RxLength should be zero on error.
 	 */
-	RESPONSECODE return_value = IFD_COMMUNICATION_ERROR;
+	RESPONSECODE return_value = IFD_ERROR_NOT_SUPPORTED;
 	int reader_index;
+	_ccid_descriptor *ccid_descriptor;
 
 	reader_index = LunToReaderIndex(Lun);
 	if ((-1 == reader_index) || (NULL == pdwBytesReturned))
-		return return_value;
+		return IFD_COMMUNICATION_ERROR;
+
+	ccid_descriptor = get_ccid_descriptor(reader_index);
 
 	DEBUG_INFO4("ControlCode: 0x%X, %s (lun: %X)", dwControlCode,
 		CcidSlots[reader_index].readerName, Lun);
@@ -1215,7 +1328,29 @@ EXTERNAL RESPONSECODE IFDHControl(DWORD Lun, DWORD dwControlCode,
 
 	if (IOCTL_SMARTCARD_VENDOR_IFD_EXCHANGE == dwControlCode)
 	{
-		if (FALSE == (DriverOptions & DRIVER_OPTION_CCID_EXCHANGE_AUTHORIZED))
+		int allowed = (DriverOptions & DRIVER_OPTION_CCID_EXCHANGE_AUTHORIZED);
+		int readerID = ccid_descriptor -> readerID;
+
+		if (VENDOR_GEMALTO == GET_VENDOR(readerID))
+		{
+			unsigned char switch_interface[] = { 0x52, 0xF8, 0x04, 0x01, 0x00 };
+
+			/* get firmware version escape command */
+			if ((1 == TxLength) && (0x02 == TxBuffer[0]))
+				allowed = TRUE;
+
+			/* switch interface escape command on the GemProx DU
+			 * the next byte in the command is the interface:
+			 * 0x01 switch to contactless interface
+			 * 0x02 switch to contact interface
+			 */
+			if ((GEMALTOPROXDU == readerID)
+				&& (6 == TxLength)
+				&& (0 == memcmp(TxBuffer, switch_interface, sizeof(switch_interface))))
+				allowed = TRUE;
+		}
+
+		if (!allowed)
 		{
 			DEBUG_INFO("ifd exchange (Escape command) not allowed");
 			return_value = IFD_COMMUNICATION_ERROR;
@@ -1231,21 +1366,21 @@ EXTERNAL RESPONSECODE IFDHControl(DWORD Lun, DWORD dwControlCode,
 		}
 	}
 
-	/* Implement the PC/SC v2.02.05 Part 10 IOCTL mechanism */
+	/* Implement the PC/SC v2.02.07 Part 10 IOCTL mechanism */
 
 	/* Query for features */
 	if (CM_IOCTL_GET_FEATURE_REQUEST == dwControlCode)
 	{
 		unsigned int iBytesReturned = 0;
 		PCSC_TLV_STRUCTURE *pcsc_tlv = (PCSC_TLV_STRUCTURE *)RxBuffer;
+		int readerID = ccid_descriptor -> readerID;
 
-		/* we need room for up to for records */
-		if (RxLength < 4 * sizeof(PCSC_TLV_STRUCTURE))
-			return IFD_COMMUNICATION_ERROR;
+		/* we need room for up to five records */
+		if (RxLength < 5 * sizeof(PCSC_TLV_STRUCTURE))
+			return IFD_ERROR_INSUFFICIENT_BUFFER;
 
 		/* We can only support direct verify and/or modify currently */
-		if (get_ccid_descriptor(reader_index) -> bPINSupport
-			& CCID_CLASS_PIN_VERIFY)
+		if (ccid_descriptor -> bPINSupport & CCID_CLASS_PIN_VERIFY)
 		{
 			pcsc_tlv -> tag = FEATURE_VERIFY_PIN_DIRECT;
 			pcsc_tlv -> length = 0x04; /* always 0x04 */
@@ -1255,8 +1390,7 @@ EXTERNAL RESPONSECODE IFDHControl(DWORD Lun, DWORD dwControlCode,
 			iBytesReturned += sizeof(PCSC_TLV_STRUCTURE);
 		}
 
-		if (get_ccid_descriptor(reader_index) -> bPINSupport
-			& CCID_CLASS_PIN_MODIFY)
+		if (ccid_descriptor -> bPINSupport & CCID_CLASS_PIN_MODIFY)
 		{
 			pcsc_tlv -> tag = FEATURE_MODIFY_PIN_DIRECT;
 			pcsc_tlv -> length = 0x04; /* always 0x04 */
@@ -1266,7 +1400,6 @@ EXTERNAL RESPONSECODE IFDHControl(DWORD Lun, DWORD dwControlCode,
 			iBytesReturned += sizeof(PCSC_TLV_STRUCTURE);
 		}
 
-#ifdef FEATURE_IFD_PIN_PROPERTIES
 		/* We can always forward wLcdLayout */
 		pcsc_tlv -> tag = FEATURE_IFD_PIN_PROPERTIES;
 		pcsc_tlv -> length = 0x04; /* always 0x04 */
@@ -1274,42 +1407,108 @@ EXTERNAL RESPONSECODE IFDHControl(DWORD Lun, DWORD dwControlCode,
 
 		pcsc_tlv++;
 		iBytesReturned += sizeof(PCSC_TLV_STRUCTURE);
-#endif
 
-		if (KOBIL_TRIBANK == get_ccid_descriptor(reader_index) -> readerID)
+		if ((KOBIL_TRIBANK == readerID)
+			|| (KOBIL_MIDENTITY_VISUAL == readerID))
 		{
-			pcsc_tlv -> tag = FEATURE_MCT_READERDIRECT;
+			pcsc_tlv -> tag = FEATURE_MCT_READER_DIRECT;
 			pcsc_tlv -> length = 0x04; /* always 0x04 */
-			pcsc_tlv -> value = htonl(IOCTL_FEATURE_MCT_READERDIRECT);
+			pcsc_tlv -> value = htonl(IOCTL_FEATURE_MCT_READER_DIRECT);
 
 			pcsc_tlv++;
 			iBytesReturned += sizeof(PCSC_TLV_STRUCTURE);
 		}
 
+		pcsc_tlv -> tag = FEATURE_GET_TLV_PROPERTIES;
+		pcsc_tlv -> length = 0x04; /* always 0x04 */
+		pcsc_tlv -> value = htonl(IOCTL_FEATURE_GET_TLV_PROPERTIES);
+		pcsc_tlv++;
+		iBytesReturned += sizeof(PCSC_TLV_STRUCTURE);
+
 		*pdwBytesReturned = iBytesReturned;
 		return_value = IFD_SUCCESS;
 	}
 
-#ifdef FEATURE_IFD_PIN_PROPERTIES
 	/* Get PIN handling capabilities */
 	if (IOCTL_FEATURE_IFD_PIN_PROPERTIES == dwControlCode)
 	{
 		PIN_PROPERTIES_STRUCTURE *caps = (PIN_PROPERTIES_STRUCTURE *)RxBuffer;
 
 		if (RxLength < sizeof(PIN_PROPERTIES_STRUCTURE))
-			return IFD_COMMUNICATION_ERROR;
+			return IFD_ERROR_INSUFFICIENT_BUFFER;
 
 		/* Only give the LCD size for now */
-		caps -> wLcdLayout = get_ccid_descriptor(reader_index) -> wLcdLayout;
-		caps -> wLcdMaxCharacters = 0x0000;
-		caps -> wLcdMaxLines = 0x0000; 
+		caps -> wLcdLayout = ccid_descriptor -> wLcdLayout;
 		caps -> bEntryValidationCondition = 0x07; /* Default */
 		caps -> bTimeOut2 = 0x00; /* We do not distinguish bTimeOut from TimeOut2 */
 
 		*pdwBytesReturned = sizeof(*caps);
 		return_value = IFD_SUCCESS;
 	}
-#endif
+
+	/* Reader features */
+	if (IOCTL_FEATURE_GET_TLV_PROPERTIES == dwControlCode)
+	{
+		int p = 0;
+		int tmp;
+
+		/* wLcdLayout */
+		RxBuffer[p++] = PCSCv2_PART10_PROPERTY_wLcdLayout;	/* tag */
+		RxBuffer[p++] = 2;	/* length */
+		tmp = ccid_descriptor -> wLcdLayout;
+		RxBuffer[p++] = tmp & 0xFF;	/* value in little endian order */
+		RxBuffer[p++] = (tmp >> 8) & 0xFF;
+
+		/* bTimeOut2 */
+		RxBuffer[p++] = PCSCv2_PART10_PROPERTY_bTimeOut2;
+		RxBuffer[p++] = 1;	/* length */
+		/* IFD does not distinguish bTimeOut from bTimeOut2 */
+		RxBuffer[p++] = 0x00;
+
+		/* sFirmwareID */
+		if (VENDOR_GEMALTO == GET_VENDOR(ccid_descriptor -> readerID))
+		{
+			unsigned char firmware[256];
+			const unsigned char cmd[] = { 0x02 };
+			RESPONSECODE ret;
+			unsigned int len;
+
+			len = sizeof(firmware);
+			ret = CmdEscape(reader_index, cmd, sizeof(cmd), firmware, &len);
+
+			if (IFD_SUCCESS == ret)
+			{
+				RxBuffer[p++] = PCSCv2_PART10_PROPERTY_sFirmwareID;
+				RxBuffer[p++] = len;
+				memcpy(&RxBuffer[p], firmware, len);
+				p += len;
+			}
+		}
+
+		/* Gemalto PC Pinpad V1 */
+		if ((GEMPCPINPAD == ccid_descriptor -> readerID)
+			/* Covadis Véga-Alpha */
+			|| (VEGAALPHA == ccid_descriptor->readerID))
+		{
+			/* bMinPINSize */
+			RxBuffer[p++] = PCSCv2_PART10_PROPERTY_bMinPINSize;
+			RxBuffer[p++] = 1;	/* length */
+			RxBuffer[p++] = 4;	/* min PIN size */
+
+			/* bMaxPINSize */
+			RxBuffer[p++] = PCSCv2_PART10_PROPERTY_bMaxPINSize;
+			RxBuffer[p++] = 1;	/* length */
+			RxBuffer[p++] = 8;	/* max PIN size */
+
+			/* bEntryValidationCondition */
+			RxBuffer[p++] = PCSCv2_PART10_PROPERTY_bEntryValidationCondition;
+			RxBuffer[p++] = 1;	/* length */
+			RxBuffer[p++] = 0x02;	/* validation key pressed */
+		}
+
+		*pdwBytesReturned = p;
+		return_value = IFD_SUCCESS;
+	}
 
 	/* Verify a PIN, plain CCID */
 	if (IOCTL_FEATURE_VERIFY_PIN_DIRECT == dwControlCode)
@@ -1334,7 +1533,7 @@ EXTERNAL RESPONSECODE IFDHControl(DWORD Lun, DWORD dwControlCode,
 	}
 
 	/* MCT: Multifunctional Card Terminal */
-	if (IOCTL_FEATURE_MCT_READERDIRECT == dwControlCode)
+	if (IOCTL_FEATURE_MCT_READER_DIRECT == dwControlCode)
 	{
 		if ( (TxBuffer[0] != 0x20)	/* CLA */
 			|| ((TxBuffer[1] & 0xF0) != 0x70)	/* INS */
@@ -1525,9 +1724,10 @@ CcidDesc *get_ccid_slot(unsigned int reader_index)
 
 void init_driver(void)
 {
-	char keyValue[TOKEN_MAX_VALUE_SIZE];
 	char infofile[FILENAME_MAX];
 	char *e;
+	int rv;
+	list_t plist, *values;
 
 	DEBUG_INFO("Driver version: " VERSION);
 
@@ -1535,11 +1735,16 @@ void init_driver(void)
 	(void)snprintf(infofile, sizeof(infofile), "%s/%s/Contents/Info.plist",
 		PCSCLITE_HP_DROPDIR, BUNDLE);
 
+	rv = bundleParse(infofile, &plist);
+	if (rv)
+		return;
+
 	/* Log level */
-	if (0 == LTPBundleFindValueWithKey(infofile, "ifdLogLevel", keyValue, 0))
+	rv = LTPBundleFindValueWithKey(&plist, "ifdLogLevel", &values);
+	if (0 == rv)
 	{
 		/* convert from hex or dec or octal */
-		LogLevel = strtoul(keyValue, NULL, 0);
+		LogLevel = strtoul(list_get_at(values, 0), NULL, 0);
 
 		/* print the log level used */
 		DEBUG_INFO2("LogLevel: 0x%.4X", LogLevel);
@@ -1554,17 +1759,20 @@ void init_driver(void)
 		/* print the log level used */
 		DEBUG_INFO2("LogLevel from LIBCCID_ifdLogLevel: 0x%.4X", LogLevel);
 	}
-	
+
 	/* Driver options */
-	if (0 == LTPBundleFindValueWithKey(infofile, "ifdDriverOptions", keyValue, 0))
+	rv = LTPBundleFindValueWithKey(&plist, "ifdDriverOptions", &values);
+	if (0 == rv)
 	{
 		/* convert from hex or dec or octal */
-		DriverOptions = strtoul(keyValue, NULL, 0);
+		DriverOptions = strtoul(list_get_at(values, 0), NULL, 0);
 
 		/* print the log level used */
 		DEBUG_INFO2("DriverOptions: 0x%.4X", DriverOptions);
 	}
 
+	bundleRelease(&plist);
+
 	/* get the voltage parameter */
 	switch ((DriverOptions >> 4) & 0x03)
 	{
@@ -1736,16 +1944,11 @@ static unsigned int T0_card_timeout(double f, double d, int TC1, int TC2,
 
 	/* ISO in */
 	t  = 261 * EGT + (3 + 3) * WWT;
-	/* Convert from milliseonds to seconds rouned to the upper value
-	 * use +1 instead of ceil() to round up to the nearest integer
-	 * so we can avoid a dependency on the math library */
-	t = t/1000 +1;
 	if (timeout < t)
 		timeout = t;
 
 	/* ISO out */
 	t = 5 * EGT + (1 + 259) * WWT;
-	t = t/1000 +1;
 	if (timeout < t)
 		timeout = t;
 
@@ -1793,11 +1996,46 @@ static unsigned int T1_card_timeout(double f, double d, int TC1,
 
 	timeout = 260*EGT + BWT + 260*CWT;
 
-	/* Convert from milliseonds to seconds rounded to the upper value
-	 * we use +1 instead of ceil() to round up to the nearest greater integer
-	 * so we can avoid a dependency on the math library */
-	timeout = timeout/1000 +1;
+	/* This is the card/reader timeout.  Add 1 second for the libusb
+	 * timeout so we get the error from the reader. */
+	timeout += 1000;
 
 	return timeout;
 } /* T1_card_timeout  */
 
+
+static int get_IFSC(ATR_t *atr, int *idx)
+{
+	int i, ifsc, protocol = -1;
+
+	/* default return values */
+	ifsc = -1;
+	*idx = -1;
+
+	for (i=0; i<ATR_MAX_PROTOCOLS; i++)
+	{
+		/* protocol T=? */
+		if (atr->ib[i][ATR_INTERFACE_BYTE_TD].present)
+			protocol = atr->ib[i][ATR_INTERFACE_BYTE_TD].value & 0x0F;
+
+		/* TAi (i>2) present and protocol=1 => IFSC */
+		if (i >= 2 && protocol == 1
+			&& atr->ib[i][ATR_INTERFACE_BYTE_TA].present)
+		{
+			ifsc = atr->ib[i][ATR_INTERFACE_BYTE_TA].value;
+			*idx = i+1;
+			/* only the first TAi (i>2) must be used */
+			break;
+		}
+	}
+
+	if (ifsc > 254)
+	{
+		/* 0xFF is not a valid value for IFSC */
+		DEBUG_INFO2("Non ISO IFSC: 0x%X", ifsc);
+		ifsc = 254;
+	}
+
+	return ifsc;
+} /* get_IFSC */
+
